#include <stdio.h>
#include "ABBA_layout.h"

/* Virtual tables forward declaration */

static const VTABLE* const ABBA_vtbl;
static const VTABLE* const ABBA_BA_vtbl;
static const VTABLE* const ABBA_A_vtbl;
static const VTABLE* const ABBA_B_vtbl;
static const VTABLE* const ABBA_vtt[6];

/*
 * Constructors and destructors
 */

void ABBA_ctor(register ABBA* const this,
	       long a, long b, long ab, long ba, long abba)
{
  A_ctor(BASE(this,A), a);              /* compiler + user */
  B_ctor(BASE(this,B), b);              /* compiler + user */
  AB_ctor_((AB*)this, ABBA_vtt, a, b, ab); /* compiler + user */
  BA_ctor_(BASE(this,BA), ABBA_vtt+3, a, b, ba); /* compiler + user */
  this->vptr = ABBA_vtbl;               /* compiler */
  this->BA_vptr = ABBA_BA_vtbl;         /* compiler */
  this->A_vptr = ABBA_A_vtbl;           /* compiler */
  this->B_vptr = ABBA_B_vtbl;           /* compiler */
  this->abba = abba;                    /* user     */

  print_layout("ABBA", "ctor", this, sizeof *this);
}

void ABBA_dtor(register ABBA* const this)
{
  this->abba = 0;                       /* user     */

  print_layout("ABBA", "dtor", this, sizeof *this);

  BA_dtor_(BASE(this,BA), ABBA_vtt+3);  /* compiler */
  AB_dtor_((AB*)this, ABBA_vtt);        /* compiler */
  B_dtor(BASE(this,B));                 /* compiler */
  A_dtor(BASE(this,A));                 /* compiler */
}

/*
 * Member functions
 */

ABBA* ABBA_g(register ABBA* const this)
{
  print_vfunc("ABBA", "g", this,
	      "a = %ld, b = %ld, ab = %ld, ba = %ld, abba = %ld",
	      VBASE(this,ABBA_A)->a, VBASE(this,ABBA_B)->b,
	      this->ab, this->ba, this->abba);
  return this;
}

ABBA* ABBA_h(register ABBA* const this)
{
  print_vfunc("ABBA", "h", this,
	      "a = %ld, b = %ld, ab = %ld, ba = %ld, abba = %ld",
	      VBASE(this,ABBA_A)->a, VBASE(this,ABBA_B)->b,
	      this->ab, this->ba, this->abba);
  return this;
}


/* -- -- -- -- -- Compiler extra information -- -- -- -- -- */

/*
 * Runtime Type Information (RTTI)
 */

static const struct base_info ABBA_base[] = {
  { AB_info, 0               , protected_base             },
  { BA_info, VOFFSET(ABBA,BA), public_base                },
  { A_info , VOFFSET(ABBA,A) , public_base | virtual_base },
  { B_info , VOFFSET(ABBA,B) , public_base | virtual_base }
};

const type_info ABBA_info[] = {{
  "ABBA", ABBA_base, NBASE(ABBA)
}};

/*
 * Virtual tables
 */

static const VTABLE ABBA_vtbl_[] = {
  VOFFSET(ABBA,B),                      /* ABBA->B                */
  VOFFSET(ABBA,A),                      /* ABBA->A                */
  0,                                    /* ABBA->ABBA             */
  (VTABLE)ABBA_info,                    /* type_info              */
  (VTABLE)ABBA_dtor,                    /* dtor                   */
  (VTABLE)BA_f_ABBA,                    /* BA_f + ABBA->BA        */
  (VTABLE)ABBA_g,                       /* g                      */
  (VTABLE)ABBA_h                        /* h                      */
};
static const VTABLE* const ABBA_vtbl = ABBA_vtbl_ + VSTART(ABBA);

static const VTABLE ABBA_BA_vtbl_[] = {
  VOFFSET2(ABBA,BA,A),                  /* BA->A                  */
  VOFFSET2(ABBA,BA,B),                  /* BA->B                  */
  -VOFFSET(ABBA,BA),                    /* BA->ABBA               */
  (VTABLE)ABBA_info,                    /* type_info              */
  (VTABLE)ABBA_dtor_top,                /* dtor + BA->ABBA        */
  (VTABLE)ABBA_g_BA_B,                  /* g + BA->ABBA + ABBA->B */
  (VTABLE)BA_f                          /* BA_f                   */
};
static const VTABLE* const ABBA_BA_vtbl = ABBA_BA_vtbl_ + VSTART(BA);

static const VTABLE ABBA_A_vtbl_[] = {
  -VOFFSET(ABBA,A),                     /* A->ABBA                */
  (VTABLE)ABBA_info,                    /* type_info              */
  (VTABLE)ABBA_dtor_top,                /* dtor + A->ABBA         */
  (VTABLE)BA_f_A                        /* BA_f + A->BA           */
};
static const VTABLE* const ABBA_A_vtbl = ABBA_A_vtbl_ + VSTART(NO_VBASE);

static const VTABLE ABBA_B_vtbl_[] = {
  -VOFFSET(ABBA,B),                     /* B->ABBA                */
  (VTABLE)ABBA_info,                    /* type_info              */
  (VTABLE)ABBA_dtor_top,                /* dtor + B->ABBA         */
  (VTABLE)ABBA_g_B_B                    /* g + B->ABBA + ABBA->B  */
};
static const VTABLE* const ABBA_B_vtbl = ABBA_B_vtbl_ + VSTART(NO_VBASE);

/* ----- Partial ctor and dtor vtables ----- */

static const VTABLE AB_vtbl_ABBA_[] = {
  VOFFSET(ABBA,B),                      /* AB->B                  */
  VOFFSET(ABBA,A),                      /* AB->A                  */
  0,                                    /* AB->AB                 */
  (VTABLE)AB_info,                      /* type_info              */
  (VTABLE)AB_dtor,                      /* dtor                   */
  (VTABLE)A_f_AB ,                      /* A_f + AB->A            */
  (VTABLE)AB_g,                         /* g                      */
  (VTABLE)AB_h                          /* h                      */
};

static const VTABLE AB_A_vtbl_ABBA_[] = {
  -VOFFSET(ABBA,A),                     /* A->AB                  */
  (VTABLE)AB_info,                      /* type info              */
  (VTABLE)AB_dtor_top,                  /* dtor + A->AB           */
  (VTABLE)A_f                           /* A_f                    */
}; 

static const VTABLE AB_B_vtbl_ABBA_[] = {
  -VOFFSET(ABBA,B),                     /* B->AB                  */
  (VTABLE)AB_info,                      /* type info              */
  (VTABLE)AB_dtor_top,                  /* dtor + B->AB           */
  (VTABLE)AB_g_B_B,                     /* g + B->AB + AB->B      */
};

static const VTABLE BA_vtbl_ABBA_[] = {
  VOFFSET2(ABBA,BA,A),                  /* BA->A                  */
  VOFFSET2(ABBA,BA,B),                  /* BA->B                  */
  0,                                    /* top offset             */
  (VTABLE)BA_info,                      /* type_info              */
  (VTABLE)BA_dtor,                      /* dtor                   */
  (VTABLE)B_g_BA,                       /* B_g + BA->B            */
  (VTABLE)BA_f                          /* f                      */
};

static const VTABLE BA_B_vtbl_ABBA_[] = {
  -VOFFSET2(ABBA,BA,B),                 /* B->BA                  */
  (VTABLE)BA_info,                      /* type info              */
  (VTABLE)BA_dtor_top,                  /* dtor + B->BA           */
  (VTABLE)B_g                           /* B_g                    */
};

static const VTABLE BA_A_vtbl_ABBA_[] = {
  -VOFFSET2(ABBA,BA,A),                 /* A->BA                  */
  (VTABLE)BA_info,                      /* type info              */
  (VTABLE)BA_dtor,                      /* dtor + A->BA           */
  (VTABLE)BA_f_A                        /* f + A->BA              */
};

static const VTABLE* const ABBA_vtt[] = {
  AB_vtbl_ABBA_   + VSTART(AB),
  AB_A_vtbl_ABBA_ + VSTART(NO_VBASE),
  AB_B_vtbl_ABBA_ + VSTART(NO_VBASE),
  BA_vtbl_ABBA_   + VSTART(BA),
  BA_B_vtbl_ABBA_ + VSTART(NO_VBASE),
  BA_A_vtbl_ABBA_ + VSTART(NO_VBASE)
};

/*
 * Partial constructors and destructors
 */

void ABBA_ctor_(register ABBA* const this, const VTABLE* const vtt[],
		long a, long b, long ab, long ba, long abba)
{
  AB_ctor_((AB*)this, vtt+4, a, b, ab);
  BA_ctor_(BASE(this,BA), vtt+7, a, b, ba);
  this->vptr = vtt[0];
  this->BA_vptr = vtt[1];
  VBASE(this,ABBA_A)->vptr = vtt[2];
  VBASE(this,ABBA_B)->vptr = vtt[3];
  this->abba = abba;

  print_layout("ABBA", "ctor_", this, sizeof *this - sizeof(A)-sizeof(B));
}

void ABBA_dtor_(register ABBA* const this, const VTABLE* const vtt[])
{
  this->vptr = vtt[0];
  this->BA_vptr = vtt[1];
  VBASE(this,ABBA_A)->vptr = vtt[2];
  VBASE(this,ABBA_B)->vptr = vtt[3];
  this->abba = 0;

  print_layout("ABBA", "dtor_", this, sizeof *this - sizeof(A)-sizeof(B));

  BA_dtor_(BASE(this,BA), vtt+7);
  AB_dtor_((AB*)this, vtt+4);
}

/*
 * Member functions thunks
 */

void ABBA_dtor_top(register ABBA* const this)
{
  print_thunk("ABBA", "dtor_top", this, this->vptr[INDEX(TOP)], 0);

  ABBA_dtor(VTOP(this));
}

void BA_f_ABBA(register ABBA* const this)
{
  print_thunk("BA", "f_ABBA", this, VOFFSET(ABBA,BA), 0);

  BA_f(BASE(this,BA));
}

B* ABBA_g_BA_B(register BA* const this)
{
  ABBA *tmp;

  print_thunk("ABBA", "g_BA_B", this, -VOFFSET(ABBA,BA), 0);

  tmp = ABBA_g(DERIVED(this,BA,ABBA));

  print_thunk("ABBA", "g_BA_B", this, tmp->vptr[INDEX(ABBA_B)], 1);

  return VBASE(tmp,ABBA_B);
}

B* ABBA_g_B_B(register B* const this)
{
  ABBA *tmp;

  print_thunk("ABBA", "g_B_B", this, this->vptr[INDEX(TOP)], 0);

  tmp = ABBA_g(VTOP(this));

  print_thunk("ABBA", "g_B_B", this, tmp->vptr[INDEX(ABBA_B)], 1);

  return VBASE(tmp,ABBA_B);
}
