#ifndef ABBA_LAYOUT_H
#define ABBA_LAYOUT_H

#include "AB_layout.h"
#include "BA_layout.h"

/* -----
 * NOTE: Layout of ABBA
 * -----
 *
 * struct ABBA : protected AB, BA {
 *   ABBA(long a, long b, long ab, long ba, long abba_);
 *   ~ABBA();
 *   ABBA* g();
 *   ABBA* h();
 * protected:
 *   long abba;
 * };
 *
 * -----
 */

/*
 * Instance layout
 */

typedef struct {
  const VTABLE *vptr;                   /* compiler    */
  long ab;                              /* AB member   */
  const VTABLE *BA_vptr;                /* compiler    */
  long ba;                              /* BA member   */
  long abba;                            /* ABBA member */
  const VTABLE *A_vptr;                 /* compiler    */
  long a;                               /* A member    */
  const VTABLE *B_vptr;                 /* compiler    */
  long b;                               /* B member    */
} ABBA;

/*
 * Member functions
 */

void  ABBA_ctor(ABBA *this, long a, long b, long ab, long ba, long abba);
void  ABBA_dtor(ABBA *this);
ABBA* ABBA_g   (ABBA *this);
ABBA* ABBA_h   (ABBA *this);


/* -- -- -- -- -- Compiler extra information -- -- -- -- -- */

/*
 * Virtual bases
 */

enum {
  INDEX(ABBA_A   ) = INDEX(AB_A),
  INDEX(ABBA_B   ) = INDEX(AB_B),
  INDEX(ABBA_LAST) = INDEX(ABBA_B)
};

typedef A ABBA_A;
typedef B ABBA_B;

/*
 * Virtual member functions
 */

enum {
  INDEX(ABBA_dtor) = INDEX(AB_dtor),
  INDEX(ABBA_f   ) = INDEX(AB_f),
  INDEX(ABBA_g   ) = INDEX(AB_g),
  INDEX(ABBA_h   ) = AFTER_VFUNC(ABBA_g)
};

typedef void  (TYPE(ABBA_dtor))(ABBA *this);
typedef void  (TYPE(ABBA_f   ))(const ABBA *this);
typedef ABBA* (TYPE(ABBA_g   ))(ABBA *this);
typedef ABBA* (TYPE(ABBA_h   ))(ABBA *this);


/* -- -- -- -- -- Compiler exported symbols -- -- -- -- -- */

/*
 * Runtime Type Information (RTTI)
 */

extern const type_info ABBA_info[];

/*
 * Partial constructors and destructors
 */

void ABBA_ctor_(ABBA *this, const VTABLE* const vtt[],
		long a, long b, long ab, long ba, long abba);
void ABBA_dtor_(ABBA *this, const VTABLE* const vtt[]);

/*
 * Member functions thunks
 */

void ABBA_dtor_top(ABBA *this);
void BA_f_ABBA    (ABBA *this);
B*   ABBA_g_BA_B  (BA   *this);
B*   ABBA_g_B_B   (B    *this);

#endif  /* ABBA_LAYOUT_H */
