#include <stdio.h>
#include "AB_layout.h"

/* Virtual tables forward declaration */

static const VTABLE* const AB_vtbl;
static const VTABLE* const AB_A_vtbl;
static const VTABLE* const AB_B_vtbl;

/*
 * Constructors and destructors
 */

void AB_ctor(register AB* const this, long a, long b, long ab)
{
  A_ctor(BASE(this,A), a);              /* compiler + user */
  B_ctor(BASE(this,B), b);              /* compiler + user */
  this->vptr = AB_vtbl;                 /* compiler        */
  this->A_vptr = AB_A_vtbl;             /* compiler        */
  this->B_vptr = AB_B_vtbl;             /* compiler        */
  this->ab = ab;                        /* user            */

  print_layout("AB", "ctor", this, sizeof *this);
}

void AB_dtor(register AB* const this)
{
  this->vptr = AB_vtbl;                 /* compiler */
  this->A_vptr = AB_A_vtbl;             /* compiler */
  this->B_vptr = AB_B_vtbl;             /* compiler */
  this->ab = 0;                         /* user     */

  print_layout("AB", "dtor", this, sizeof *this);

  B_dtor(BASE(this,B));                 /* compiler */
  A_dtor(BASE(this,A));                 /* compiler */
}

/*
 * Member functions
 */

AB* AB_g(register AB* const this)
{
  print_vfunc("AB", "g", this, 
	      "a = %ld, b = %ld, ab = %ld",
	      VBASE(this,AB_A)->a, VBASE(this,AB_B)->b, this->ab);
  return this;
}

AB* AB_h(register AB* const this)
{
  print_vfunc("AB", "h", this, 
	      "a = %ld, b = %ld, ab = %ld",
	      VBASE(this,AB_A)->a, VBASE(this,AB_B)->b, this->ab);
  return this;
}


/* -- -- -- -- -- Compiler extra information -- -- -- -- -- */

/*
 * Runtime Type Information (RTTI)
 */

static const struct base_info AB_base[] = {
  { A_info, VOFFSET(AB,A), public_base | virtual_base },
  { B_info, VOFFSET(AB,B), public_base | virtual_base }
};

const type_info AB_info[] = {{
  "AB", AB_base, NBASE(AB)
}};

/*
 * Virtual tables
 */

static const VTABLE AB_vtbl_[] = {
  VOFFSET(AB,B),                        /* AB->B             */
  VOFFSET(AB,A),                        /* AB->A             */
  0,                                    /* AB->AB            */
  (VTABLE)AB_info,                      /* type_info         */
  (VTABLE)AB_dtor,                      /* dtor              */
  (VTABLE)A_f_AB ,                      /* A_f + AB->A       */
  (VTABLE)AB_g,                         /* g                 */
  (VTABLE)AB_h                          /* h                 */
};
static const VTABLE* const AB_vtbl = AB_vtbl_ + VSTART(AB);

static const VTABLE AB_A_vtbl_[] = {
  -VOFFSET(AB,A),                       /* A->AB             */
  (VTABLE)AB_info,                      /* type_info         */
  (VTABLE)AB_dtor_top,                  /* dtor + A->AB      */
  (VTABLE)A_f                           /* A_f               */
};
static const VTABLE* const AB_A_vtbl = AB_A_vtbl_ + VSTART(NO_VBASE);

static const VTABLE AB_B_vtbl_[] = {
  -VOFFSET(AB,B),                       /* B->AB             */
  (VTABLE)AB_info,                      /* type_info         */
  (VTABLE)AB_dtor_top,                  /* dtor + B->AB      */
  (VTABLE)AB_g_B_B                      /* g + B->AB + B<-AB */
};
static const VTABLE* const AB_B_vtbl = AB_B_vtbl_ + VSTART(NO_VBASE);

/*
 * Partial constructors and destructors
 */

void AB_ctor_(register AB* const this, const VTABLE* const vtt[],
	      long a, long b, long ab)
{
  this->vptr = vtt[0];
  VBASE(this,AB_A)->vptr = vtt[1];
  VBASE(this,AB_B)->vptr = vtt[2];
  this->ab = ab;
  UNUSED2(a,b);

  print_layout("AB", "ctor_", this, sizeof *this - sizeof(A)-sizeof(B));
}

void AB_dtor_(register AB* const this, const VTABLE* const vtt[])
{
  this->vptr = vtt[0];
  VBASE(this,AB_A)->vptr = vtt[1];
  VBASE(this,AB_B)->vptr = vtt[2];
  this->ab = 0;

  print_layout("AB", "dtor_", this, sizeof *this - sizeof(A)-sizeof(B));
}

/*
 * Member functions thunks
 */

void AB_dtor_top(register AB* const this)
{
  print_thunk("AB", "dtor_top", this, this->vptr[INDEX(TOP)], 0);

  AB_dtor(VTOP(this));
}

void A_f_AB(register AB* const this)
{
  print_thunk("A", "f_AB", this, this->vptr[INDEX(AB_A)], 0);

  A_f(VBASE(this,AB_A));
}

B* AB_g_B_B(register B* const this)
{
  AB *tmp;

  print_thunk("AB", "g_B_B", this, this->vptr[INDEX(TOP)], 0);

  tmp = AB_g(VTOP(this));

  print_thunk("AB", "g_B_B", this, tmp->vptr[INDEX(AB_B)], 1);

  return VBASE(this,AB_B);
}
