#ifndef AB_LAYOUT_H
#define AB_LAYOUT_H

#include "A_layout.h"
#include "B_layout.h"

/* -----
 * NOTE: Layout of AB
 * -----
 *
 * struct AB : virtual A, virtual B {
 *   AB(long a, long b, long ab_);
 *   ~AB();
 *   AB* g();
 *   virtual AB* h();
 * protected:
 *   long ab;
 * };
 *
 * -----
 */

/*
 * Instance layout
 */

typedef struct {
  const VTABLE *vptr;                   /* compiler  */
  long ab;                              /* AB member */
  const VTABLE *A_vptr;                 /* compiler  */
  long a;                               /* A member  */
  const VTABLE *B_vptr;                 /* compiler  */
  long b;                               /* B member  */
} AB;

/*
 * Member functions
 */

void AB_ctor(AB *this, long a, long b, long ab);
void AB_dtor(AB *this);
AB*  AB_g   (AB *this);
AB*  AB_h   (AB *this);


/* -- -- -- -- -- Compiler extra information -- -- -- -- -- */

/*
 * Virtual bases
 */

enum {
  INDEX(AB_A   ) = FIRST_VBASE,
  INDEX(AB_B   ) = AFTER_VBASE(AB_A),
  INDEX(AB_LAST) = INDEX(AB_B)
};

typedef A AB_A;
typedef B AB_B;

/*
 * Virtual member functions
 */

enum {
  INDEX(AB_dtor) = INDEX(A_dtor),
  INDEX(AB_f   ) = INDEX(A_f),
  INDEX(AB_g   ) = AFTER_VFUNC(AB_f),
  INDEX(AB_h   ) = AFTER_VFUNC(AB_g)
};

typedef void (TYPE(AB_dtor))(AB*);
typedef void (TYPE(AB_f   ))(const AB*);
typedef AB*  (TYPE(AB_g   ))(AB*);
typedef AB*  (TYPE(AB_h   ))(AB*);

/* -- -- -- -- -- Compiler exported symbols -- -- -- -- -- */

/*
 * Runtime Type Information (RTTI)
 */

extern const type_info AB_info[];

/*
 * Partial constructors and destructors
 */

void AB_ctor_(AB *this, const VTABLE* const vtt[],
	      long a, long b, long ab);
void AB_dtor_(AB *this, const VTABLE* const vtt[]);

/*
 * Member functions thunks
 */

void AB_dtor_top(AB *this);
void A_f_AB     (AB *this);
B*   AB_g_B_B   (B  *this);

#endif  /* AB_LAYOUT_H */
