#ifndef A_LAYOUT_H
#define A_LAYOUT_H

#include "object_model.h"

/* -----
 * NOTE: Layout of A
 * -----
 *
 * struct A {
 *   A(long a_);
 *   virtual ~A();
 *   virtual void f() const;
 * protected:
 *   long a;
 * };
 *
 * -----
 */

/*
 * Instance layout
 */

typedef struct {
  const VTABLE *vptr;                   /* compiler */
  long a;                               /* A member */
} A;

/*
 * Member functions
 */

void A_ctor(A *this, long a);
void A_dtor(A *this);
void A_f   (const A *this);


/* -- -- -- -- -- Compiler extra information -- -- -- -- -- */

/*
 * Virtual member functions
 */

enum { 
  INDEX(A_dtor) = FIRST_VFUNC,
  INDEX(A_f   ) = AFTER_VFUNC(A_dtor)
};

typedef void (TYPE(A_dtor))(A*);
typedef void (TYPE(A_f   ))(const A*);

/* -- -- -- -- -- Compiler exported symbols -- -- -- -- -- */

/*
 * Runtime Type Information (RTTI)
 */

extern const type_info A_info[];

#endif /* A_LAYOUT_H */
