#include <stdio.h>
#include "BA_layout.h"

/* Virtual tables forward declaration */

static const VTABLE* const BA_vtbl;
static const VTABLE* const BA_B_vtbl;
static const VTABLE* const BA_A_vtbl;

/*
 * Constructors and destructors
 */

void BA_ctor(register BA* const this, long a, long b, long ba)
{
  B_ctor(BASE(this,B), b);              /* compiler + user */
  A_ctor(BASE(this,A), a);              /* compiler + user */
  this->vptr = BA_vtbl;                 /* compiler        */
  this->B_vptr = BA_B_vtbl;             /* compiler        */
  this->A_vptr = BA_A_vtbl;             /* compiler        */
  this->ba = ba;                        /* user            */

  print_layout("BA", "ctor", this, sizeof *this);
}

void BA_dtor(register BA* const this)
{
  this->vptr = BA_vtbl;                 /* compiler */
  this->B_vptr = BA_B_vtbl;             /* compiler */
  this->A_vptr = BA_A_vtbl;             /* compiler */
  this->ba = 0;                         /* user     */

  print_layout("BA", "dtor", this, sizeof *this);

  A_dtor(BASE(this,A));                 /* compiler */
  B_dtor(BASE(this,B));                 /* compiler */
}

/*
 * Member functions
 */

void BA_f(register const BA* const this)
{
  print_vfunc("BA", "f", this, 
	      "a = %ld, b = %ld, ba = %ld",
	      VBASE(this,BA_A)->a, VBASE(this,BA_B)->b, this->ba);
}


/* -- -- -- -- -- Compiler extra information -- -- -- -- -- */

/*
 * Runtime Type Information (RTTI)
 */

static const struct base_info BA_base[] = {
  { B_info, VOFFSET(BA,B), public_base | virtual_base },
  { A_info, VOFFSET(BA,A), public_base | virtual_base }
};

const type_info BA_info[] = {{
  "BA", BA_base, NBASE(BA)
}};

/*
 * Virtual tables
 */

static const VTABLE BA_vtbl_[] = {
  VOFFSET(BA,A),                        /* BA->A             */
  VOFFSET(BA,B),                        /* BA->B             */
  0,                                    /* BA->BA            */
  (VTABLE)BA_info,                      /* type_info         */
  (VTABLE)BA_dtor,                      /* dtor              */
  (VTABLE)B_g_BA ,                      /* B_g + BA->B       */
  (VTABLE)BA_f                          /* f                 */
};
static const VTABLE* const BA_vtbl = BA_vtbl_ + VSTART(BA);

static const VTABLE BA_B_vtbl_[] = {
  -VOFFSET(BA,B),                       /* B->BA             */
  (VTABLE)BA_info,                      /* type_info         */
  (VTABLE)BA_dtor_top,                  /* dtor + B->BA      */
  (VTABLE)B_g                           /* B_g               */
};
static const VTABLE* const BA_B_vtbl = BA_B_vtbl_ + VSTART(NO_VBASE);

static const VTABLE BA_A_vtbl_[] = {
  -VOFFSET(BA,A),                       /* A->BA             */
  (VTABLE)BA_info,                      /* type_info         */
  (VTABLE)BA_dtor_top,                  /* dtor + A->BA      */
  (VTABLE)BA_f_A                        /* f + A->BA         */
};
static const VTABLE* const BA_A_vtbl = BA_A_vtbl_ + VSTART(NO_VBASE);

/*
 * Partial constructors and destructors
 */

void BA_ctor_(register BA* const this, const VTABLE* const vtt[],
	      long a, long b, long ba)
{
  this->vptr = vtt[0];
  VBASE(this,BA_A)->vptr = vtt[1];
  VBASE(this,BA_B)->vptr = vtt[2];
  this->ba = ba;
  UNUSED2(a,b);

  print_layout("BA", "ctor_", this, sizeof *this - sizeof(A)-sizeof(B));
}

void BA_dtor_(register BA* const this, const VTABLE* const vtt[])
{
  this->vptr = vtt[0];
  VBASE(this,BA_A)->vptr = vtt[1];
  VBASE(this,BA_B)->vptr = vtt[2];
  this->ba = 0;

  print_layout("BA", "dtor_", this, sizeof *this - sizeof(A)-sizeof(B));
}

/*
 * Member functions thunks
 */

void BA_dtor_top(register BA* const this)
{
  print_thunk("BA", "dtor_top", this, this->vptr[INDEX(TOP)], 0);

  BA_dtor(VTOP(this));
}

B* B_g_BA(register BA* const this)
{
  print_thunk("B", "g_BA", this, this->vptr[INDEX(BA_B)], 0);

  return B_g(VBASE(this,BA_B));
}

void BA_f_A(register A* const this)
{
  print_thunk("BA", "f_A", this, this->vptr[INDEX(TOP)], 0);

  BA_f(VTOP(this));
}
