#ifndef BA_LAYOUT_H
#define BA_LAYOUT_H

#include "B_layout.h"
#include "A_layout.h"

/* -----
 * NOTE: Layout of BA
 * -----
 *
 * struct BA : virtual A, virtual B {
 *   BA(long a, long b, long ba_);
 *   ~BA();
 *   void f() const;
 * protected:
 *   long ba;
 * };
 *
 * -----
 */

/*
 * Instance layout
 */

typedef struct {
  const VTABLE *vptr;                   /* compiler  */
  long ba;                              /* BA member */
  const VTABLE *B_vptr;                 /* compiler  */
  long b;                               /* B member  */
  const VTABLE *A_vptr;                 /* compiler  */
  long a;                               /* A member  */
} BA;

/*
 * Member functions
 */

void BA_ctor(BA *this, long a, long b, long ba);
void BA_dtor(BA *this);
void BA_f   (const BA *this);


/* -- -- -- -- -- Compiler extra information -- -- -- -- -- */

/*
 * Virtual bases
 */

enum {
  INDEX(BA_B   ) = FIRST_VBASE,
  INDEX(BA_A   ) = AFTER_VBASE(BA_B),
  INDEX(BA_LAST) = INDEX(BA_A)
};

typedef B BA_B;
typedef A BA_A;

/*
 * Virtual member functions
 */

enum {
  INDEX(BA_dtor) = INDEX(B_dtor),
  INDEX(BA_g   ) = INDEX(B_g),
  INDEX(BA_f   ) = AFTER_VFUNC(BA_g)
};

typedef void (TYPE(BA_dtor))(BA*);
typedef B*   (TYPE(BA_g   ))(BA*);
typedef void (TYPE(BA_f   ))(const BA*);

/* -- -- -- -- -- Compiler exported symbols -- -- -- -- -- */

/*
 * Runtime Type Information (RTTI)
 */

extern const type_info BA_info[];

/*
 * Partial constructors and destructors
 */

void BA_ctor_(BA *this, const VTABLE* const vtt[],
	      long a, long b, long ba);
void BA_dtor_(BA *this, const VTABLE* const vtt[]);

/*
 * Member functions thunks
 */

void BA_dtor_top(BA *this);
B*   B_g_BA     (BA *this);
void BA_f_A     (A  *this);

#endif  /* BA_LAYOUT_H */
