#include <stdio.h>
#include "C_layout.h"

/* Virtual tables forward declaration */

static const VTABLE* const C_vtbl;
static const VTABLE* const C_B_vtbl;

/*
 * Constructors and destructors
 */

void C_ctor(register C* const this, long b, long c)
{
  B_ctor(BASE(this,B), b);              /* compiler + user */
  this->vptr = C_vtbl;                  /* compiler        */
  this->B_vptr = C_B_vtbl;              /* compiler        */
  this->c = c;                          /* user            */

  print_layout("C", "ctor", this, sizeof *this);
}

void C_dtor(register C* const this)
{
  this->vptr = C_vtbl;                  /* compiler */
  this->B_vptr = C_B_vtbl;              /* compiler */
  this->c = 0;                          /* user     */

  print_layout("C", "dtor", this, sizeof *this);

  B_dtor(BASE(this,B));                 /* compiler */
}

/*
 * Member functions
 */

B* C_g(register C* const this)
{
  B* const this_B = VBASE(this,C_B);

  print_vfunc("C", "g", this,
	      "b = %ld, c = %ld", this_B->b, this->c);
  return this_B;
}


/* -- -- -- -- -- Compiler extra information -- -- -- -- -- */

/*
 * Runtime Type Information (RTTI)
 */

static const struct base_info C_base[] = {
  { B_info, VOFFSET(C,B), public_base | virtual_base }
};

const type_info C_info[] = {{
  "C", C_base, NBASE(C)
}};

/*
 * Virtual tables
 */

static const VTABLE C_vtbl_[] = {
  VOFFSET(C,B),                         /* C->B             */
  0,                                    /* C->C             */
  (VTABLE)C_info,                       /* type_info        */
  (VTABLE)C_dtor,                       /* dtor             */
  (VTABLE)C_g                           /* g                */
};
static const VTABLE* const C_vtbl = C_vtbl_ + VSTART(C);

static const VTABLE C_B_vtbl_[] = {
  -VOFFSET(C,B),                        /* B->C             */
  (VTABLE)C_info,                       /* type_info        */
  (VTABLE)C_dtor_top,                   /* dtor + B->C      */
  (VTABLE)C_g_B,                        /* g + B->C         */
};
static const VTABLE* const C_B_vtbl = C_B_vtbl_ + VSTART(NO_VBASE);

/*
 * Partial constructors and destructors
 */

void C_ctor_(register C* const this, const VTABLE* const vtt[],
	     long b, long c)
{
  this->vptr = vtt[0];
  VBASE(this,C_B)->vptr = vtt[1];
  this->c = c;
  UNUSED1(b);

  print_layout("C", "ctor_", this, sizeof *this - sizeof(B));
}

void C_dtor_(register C* const this, const VTABLE* const vtt[])
{
  this->vptr = vtt[0];
  VBASE(this,C_B)->vptr = vtt[1];
  this->c = 0;

  print_layout("C", "dtor_", this, sizeof *this - sizeof(B));
}

/*
 * Member functions thunks
 */

void C_dtor_top(register B* const this)
{
  print_thunk("C", "dtor_top", this, this->vptr[INDEX(TOP)], 0);

  C_dtor(VTOP(this));
}

B* C_g_B(register B* const this)
{
  print_thunk("C", "g_B", this, this->vptr[INDEX(TOP)], 0);

  return C_g(VTOP(this));
}
