#ifndef C_LAYOUT_H
#define C_LAYOUT_H

#include "B_layout.h"

/* -----
 * NOTE: Layout of C
 * -----
 *
 * struct C : virtual B {
 *   C(long b, long c_);
 *   ~C();
 *   B* g();
 * protected:
 *   long c;
 * };
 *
 * -----
 */

/*
 * Instance layout
 */

typedef struct {
  const VTABLE *vptr;                   /* compiler */
  long c;                               /* C member */
  const VTABLE *B_vptr;                 /* compiler */
  long b;                               /* B member */
} C;

/*
 * Member functions
 */

void C_ctor(C *this, long b, long c);
void C_dtor(C *this);
B*   C_g   (C *this);


/* -- -- -- -- -- Compiler extra information -- -- -- -- -- */

/*
 * Virtual bases
 */

enum {
  INDEX(C_B   ) = FIRST_VBASE,
  INDEX(C_LAST) = INDEX(C_B)
};

typedef B C_B;

/*
 * Virtual member functions
 */

enum {
  INDEX(C_dtor) = INDEX(B_dtor),
  INDEX(C_g   ) = INDEX(B_g)
};

typedef void (TYPE(C_dtor))(C*);
typedef B*   (TYPE(C_g   ))(C*);

/* -- -- -- -- -- Compiler exported symbols -- -- -- -- -- */

/*
 * Runtime Type Information (RTTI)
 */

extern const type_info C_info[];

/*
 * Partial constructors and destructors
 */

void C_ctor_(C *this, const VTABLE* const vtt[], long b, long c);
void C_dtor_(C *this, const VTABLE* const vtt[]);

/*
 * Member functions thunks
 */

void C_dtor_top(B *this);
B*   C_g_B     (B *this);

#endif /* C_LAYOUT_H */
