#include <stdio.h>
#include "DCABBA_layout.h"

/* Virtual tables forward declaration */

static const VTABLE* const DCABBA_vtbl;
static const VTABLE* const DCABBA_ABBA_vtbl;
static const VTABLE* const DCABBA_BA_vtbl;
static const VTABLE* const DCABBA_B_vtbl;
static const VTABLE* const DCABBA_A_vtbl;
static const VTABLE* const DCABBA_vtt[12];

/*
 * Constructors and destructors
 */

void DCABBA_ctor(register DCABBA* const this,
		 long a, long b, long c, long d,
		 long ab, long ba, long abba, long dcabba)
{
  B_ctor(BASE(this,B), b);              /* compiler + user */
  A_ctor(BASE(this,A), a);              /* compiler + user */
  D_ctor((D*)&this->d, d);              /* compiler + user */
  C_ctor_((C*)this, DCABBA_vtt, b, c);  /* compiler + user */
  ABBA_ctor_(BASE(this,ABBA), DCABBA_vtt+2, a, b, ab, ba, abba); /* c+u */
  this->vptr = DCABBA_vtbl;             /* compiler */
  this->ABBA_vptr = DCABBA_ABBA_vtbl;   /* compiler */
  this->BA_vptr = DCABBA_BA_vtbl;       /* compiler */
  this->B_vptr = DCABBA_B_vtbl;         /* compiler */
  this->A_vptr = DCABBA_A_vtbl;         /* compiler */
  this->dcabba = dcabba;                /* user     */

  print_layout("DCABBA", "ctor", this, sizeof *this);
}

void DCABBA_dtor(register DCABBA* const this)
{
  this->dcabba = 0;                     /* user     */

  print_layout("DCABBA", "dtor", this, sizeof *this);

  ABBA_dtor_(BASE(this,ABBA), DCABBA_vtt+2); /* compiler */
  C_dtor_((C*)this, DCABBA_vtt);        /* compiler */
  D_dtor((D*)&this->d);                 /* compiler */
  A_dtor(BASE(this,A));                 /* compiler */
  B_dtor(BASE(this,B));                 /* compiler */
}

/*
 * Member functions
 */

void DCABBA_f(register const DCABBA* const this)
{
  print_vfunc("DCABBA", "f", this, 
	      "a = %ld, b = %ld, c = %ld, d = %ld, "
	      "ab = %ld, ba = %ld, abba = %ld, dcabba = %ld",
	      VBASE(this,DCABBA_A)->a, VBASE(this,DCABBA_B)->b, this->c,
	      this->d, this->ab, this->ba, this->abba, this->dcabba);
}


DCABBA* DCABBA_g(register DCABBA* const this)
{
  print_vfunc("DCABBA", "g", this,
	      "a = %ld, b = %ld, c = %ld, d = %ld, "
	      "ab = %ld, ba = %ld, abba = %ld, dcabba = %ld",
	      VBASE(this,DCABBA_A)->a, VBASE(this,DCABBA_B)->b, this->c,
	      this->d, this->ab, this->ba, this->abba, this->dcabba);
  return this;
}

DCABBA* DCABBA_h(register DCABBA* const this)
{
  print_vfunc("DCABBA", "h", this, 
	      "a = %ld, b = %ld, c = %ld, d = %ld, "
	      "ab = %ld, ba = %ld, abba = %ld, dcabba = %ld",
	      VBASE(this,DCABBA_A)->a, VBASE(this,DCABBA_B)->b, this->c,
	      this->d, this->ab, this->ba, this->abba, this->dcabba);
  return this;
}


/* -- -- -- -- -- Compiler extra information -- -- -- -- -- */

/*
 * Runtime Type Information (RTTI)
 */

static const struct base_info DCABBA_base[] = {
  { C_info   , 0                   , public_base                },
  { D_info   , offsetof(DCABBA,d)  , public_base                },
  { ABBA_info, VOFFSET(DCABBA,ABBA), public_base                },
  { BA_info  , VOFFSET(DCABBA,BA)  , public_base                },
  { B_info   , VOFFSET(DCABBA,B)   , public_base | virtual_base },
  { A_info   , VOFFSET(DCABBA,A)   , public_base | virtual_base }
};

const type_info DCABBA_info[] = {{
  "DCABBA", DCABBA_base, NBASE(DCABBA)
}};

/*
 * Virtual tables
 */

static const VTABLE DCABBA_vtbl_[] = {
  VOFFSET(DCABBA,A),                    /* DCABBA->A              */
  VOFFSET(DCABBA,B),                    /* DCABBA->B              */
  0,                                    /* DCABBA->DCABBA         */
  (VTABLE)DCABBA_info,                  /* type_info              */
  (VTABLE)DCABBA_dtor,                  /* dtor                   */
  (VTABLE)DCABBA_g__B,                  /* g + DCABBA->B          */
  (VTABLE)DCABBA_f,                     /* f                      */
  (VTABLE)DCABBA_g,                     /* g                      */
  (VTABLE)DCABBA_h                      /* h                      */
};
static const VTABLE* const DCABBA_vtbl = DCABBA_vtbl_ + VSTART(DCABBA);

static const VTABLE DCABBA_ABBA_vtbl_[] = {
  VOFFSET2(DCABBA,ABBA,B),              /* ABBA->B                */
  VOFFSET2(DCABBA,ABBA,A),              /* ABBA->A                */
  -VOFFSET(DCABBA,ABBA),                /* ABBA->DCABBA           */
  (VTABLE)DCABBA_info,                  /* type_info              */
  (VTABLE)DCABBA_dtor_top,              /* dtor + ABBA->DCABBA    */
  (VTABLE)DCABBA_f_ABBA,                /* f + ABBA->DCABBA       */
  (VTABLE)DCABBA_g_ABBA_ABBA,           /* g + ABBA->DCABBA + DCABBA->ABBA */
  (VTABLE)DCABBA_h_ABBA_ABBA,           /* h + ABBA->DCABBA + DCABBA->ABBA */
};
static const VTABLE* const DCABBA_ABBA_vtbl = DCABBA_ABBA_vtbl_ + VSTART(ABBA);

static const VTABLE DCABBA_BA_vtbl_[] = {
  VOFFSET2(DCABBA,BA,A),                /* BA->A                  */
  VOFFSET2(DCABBA,BA,B),                /* BA->B                  */
  -VOFFSET(DCABBA,BA),                  /* BA->DCABBA             */
  (VTABLE)DCABBA_info,                  /* type_info              */
  (VTABLE)DCABBA_dtor_top,              /* dtor + BA->DCABBA      */
  (VTABLE)DCABBA_g_BA_B,                /* g + BA->DCABBA + DCABBA->B */
  (VTABLE)DCABBA_f_BA                   /* f + BA->DCABBA         */
};
static const VTABLE* const DCABBA_BA_vtbl = DCABBA_BA_vtbl_ + VSTART(BA);

static const VTABLE DCABBA_B_vtbl_[] = {
  -VOFFSET(DCABBA,B),                   /* B->DCABBA              */
  (VTABLE)DCABBA_info,                  /* type_info              */
  (VTABLE)DCABBA_dtor_top,              /* dtor + B->DCABBA       */
  (VTABLE)DCABBA_g_B_B                  /* g + B->DCABBA + DCABBA->B */
};
static const VTABLE* const DCABBA_B_vtbl = DCABBA_B_vtbl_ + VSTART(NO_VBASE);

static const VTABLE DCABBA_A_vtbl_[] = {
  -VOFFSET(DCABBA,A),                   /* A->DCABBA              */
  (VTABLE)DCABBA_info,                  /* type_info              */
  (VTABLE)DCABBA_dtor_top,              /* dtor + A->DCABBA       */
  (VTABLE)DCABBA_f_A                    /* f + A->DCABBA          */
};
static const VTABLE* const DCABBA_A_vtbl = DCABBA_A_vtbl_ + VSTART(NO_VBASE);

/* ----- Partial ctor and dtor vtables ----- */

static const VTABLE C_vtbl_DCABBA_[] = {
  VOFFSET(DCABBA,B),                    /* C->B                   */
  0,                                    /* C->C                   */
  (VTABLE)C_info,                       /* type_info              */
  (VTABLE)C_dtor,                       /* dtor                   */
  (VTABLE)C_g                           /* g                      */
};

static const VTABLE C_B_vtbl_DCABBA_[] = {
  -VOFFSET(DCABBA,B),                   /* B->C                   */
  (VTABLE)C_info,                       /* type info              */
  (VTABLE)C_dtor_top,                   /* dtor + B->C            */
  (VTABLE)C_g_B                         /* g + B->C               */
};

static const VTABLE ABBA_vtbl_DCABBA_[] = {
  VOFFSET2(DCABBA,ABBA,B),              /* ABBA->B                */
  VOFFSET2(DCABBA,ABBA,A),              /* ABBA->A                */
  0,                                    /* ABBA->ABBA             */
  (VTABLE)ABBA_info,                    /* type_info              */
  (VTABLE)ABBA_dtor,                    /* dtor                   */
  (VTABLE)BA_f_ABBA,                    /* BA_f + ABBA->BA        */
  (VTABLE)ABBA_g,                       /* g                      */
  (VTABLE)ABBA_h                        /* h                      */
};

static const VTABLE ABBA_BA_vtbl_DCABBA_[] = {
  VOFFSET2(DCABBA,BA,A),                /* BA->A                  */
  VOFFSET2(DCABBA,BA,B),                /* BA->B                  */
  -VOFFSET2(DCABBA,ABBA,BA),            /* BA->ABBA               */
  (VTABLE)BA_info,                      /* type_info              */
  (VTABLE)BA_dtor,                      /* dtor                   */
  (VTABLE)ABBA_g_BA_B,                  /* g + BA->ABBA + ABBA->B */
  (VTABLE)BA_f                          /* f                      */
};

static const VTABLE ABBA_A_vtbl_DCABBA_[] = {
  -VOFFSET2(DCABBA,ABBA,A),             /* A->ABBA                */
  (VTABLE)ABBA_info,                    /* type info              */
  (VTABLE)ABBA_dtor_top,                /* dtor + A->ABBA         */
  (VTABLE)BA_f_A                        /* BA_f + A->BA           */
}; 

static const VTABLE ABBA_B_vtbl_DCABBA_[] = {
  -VOFFSET2(DCABBA,ABBA,B),             /* B->ABBA                */
  (VTABLE)ABBA_info,                    /* type info              */
  (VTABLE)ABBA_dtor_top,                /* dtor + B->ABBA         */
  (VTABLE)ABBA_g_B_B                    /* g + B->ABBA + ABBA->B  */
};

static const VTABLE AB_vtbl_DCABBA_[] = {
  VOFFSET2(DCABBA,ABBA,B),              /* AB->B                  */
  VOFFSET2(DCABBA,ABBA,A),              /* AB->A                  */
  0,                                    /* AB->AB                 */
  (VTABLE)AB_info,                      /* type_info              */
  (VTABLE)AB_dtor,                      /* dtor                   */
  (VTABLE)A_f_AB,                       /* A_f + AB->A            */
  (VTABLE)AB_g,                         /* g                      */
  (VTABLE)AB_h                          /* h                      */
};

static const VTABLE AB_A_vtbl_DCABBA_[] = {
  -VOFFSET2(DCABBA,ABBA,A),             /* A->AB                  */
  (VTABLE)AB_info,                      /* type info              */
  (VTABLE)AB_dtor_top,                  /* dtor + A->AB           */
  (VTABLE)A_f                           /* A_f                    */
}; 

static const VTABLE AB_B_vtbl_DCABBA_[] = {
  -VOFFSET2(DCABBA,ABBA,B),             /* B->ABBA                */
  (VTABLE)AB_info,                      /* type info              */
  (VTABLE)AB_dtor_top,                  /* dtor + B->AB           */
  (VTABLE)AB_g_B_B                      /* g + B->AB + AB->B      */
};

static const VTABLE BA_vtbl_DCABBA_[] = {
  VOFFSET2(DCABBA,BA,A),                /* BA->A                  */
  VOFFSET2(DCABBA,BA,B),                /* BA->B                  */
  0,                                    /* top offset             */
  (VTABLE)BA_info,                      /* type_info              */
  (VTABLE)BA_dtor,                      /* dtor                   */
  (VTABLE)B_g_BA,                       /* B_g + BA->B            */
  (VTABLE)BA_f                          /* f                      */
};

static const VTABLE BA_B_vtbl_DCABBA_[] = {
  -VOFFSET2(DCABBA,BA,B),               /* B->BA                  */
  (VTABLE)BA_info,                      /* type info              */
  (VTABLE)BA_dtor_top,                  /* dtor + B->BA           */
  (VTABLE)B_g                           /* B_g                    */
};

static const VTABLE BA_A_vtbl_DCABBA_[] = {
  -VOFFSET2(DCABBA,BA,A),               /* A->BA                  */
  (VTABLE)BA_info,                      /* type info              */
  (VTABLE)BA_dtor_top,                  /* dtor + A->BA           */
  (VTABLE)BA_f_A                        /* f + A->BA              */
};

static const VTABLE* const DCABBA_vtt[] = {
  C_vtbl_DCABBA_       + VSTART(C),
  C_B_vtbl_DCABBA_     + VSTART(NO_VBASE),

  ABBA_vtbl_DCABBA_    + VSTART(ABBA),
  ABBA_BA_vtbl_DCABBA_ + VSTART(BA),
  ABBA_A_vtbl_DCABBA_  + VSTART(NO_VBASE),
  ABBA_B_vtbl_DCABBA_  + VSTART(NO_VBASE),
  
  AB_vtbl_DCABBA_      + VSTART(AB),
  AB_A_vtbl_DCABBA_    + VSTART(NO_VBASE),
  AB_B_vtbl_DCABBA_    + VSTART(NO_VBASE),

  BA_vtbl_DCABBA_      + VSTART(BA),
  BA_B_vtbl_DCABBA_    + VSTART(NO_VBASE),
  BA_A_vtbl_DCABBA_    + VSTART(NO_VBASE)
};

/*
 * Partial constructors and destructors
 */

void DCABBA_ctor_(register DCABBA* const this, const VTABLE* const vtt[],
		  long a, long b, long c, long d,
		  long ab, long ba, long abba, long dcabba)
{
  D_ctor((D*)&this->d, d);
  C_ctor_((C*)this, DCABBA_vtt+5, b, c);
  ABBA_ctor_(BASE(this,ABBA), vtt+7, a, b, ab, ba, abba);
  this->vptr = vtt[0];
  this->ABBA_vptr = vtt[1];
  this->BA_vptr = vtt[2];
  VBASE(this,DCABBA_A)->vptr = vtt[3];
  VBASE(this,DCABBA_B)->vptr = vtt[4];
  this->dcabba = dcabba;

  print_layout("DCABBA", "ctor_",
	       this, sizeof *this - sizeof(A)-sizeof(B));
}

void DCABBA_dtor_(register DCABBA* const this, const VTABLE* const vtt[])
{
  this->vptr = vtt[0];
  this->ABBA_vptr = vtt[1];
  this->BA_vptr = vtt[2];
  VBASE(this,DCABBA_A)->vptr = vtt[3];
  VBASE(this,DCABBA_B)->vptr = vtt[4];
  this->dcabba = 0;

  print_layout("DCABBA", "dtor_",
	       this, sizeof *this - sizeof(A)-sizeof(B));

  ABBA_dtor_(BASE(this,ABBA), vtt+7);
  C_dtor_((C*)this, DCABBA_vtt+5);
  D_dtor((D*)&this->d);
}

/*
 * Member functions thunks
 */

void DCABBA_dtor_top(register DCABBA* const this)
{
  print_thunk("DCABBA", "dtor_top", this, this->vptr[INDEX(TOP)], 0);

  DCABBA_dtor(VTOP(this));
}

B* DCABBA_g__B(register DCABBA* const this)
{
  DCABBA *tmp = DCABBA_g(this);

  print_thunk("DCABBA", "g__B", this, tmp->vptr[INDEX(DCABBA_B)], 1);

  return VBASE(this,DCABBA_B);
}

void DCABBA_f_ABBA(register ABBA* const this)
{
  print_thunk("DCABBA", "f_ABBA", this, -VOFFSET(DCABBA,ABBA), 1);

  DCABBA_f(DERIVED(this,ABBA,DCABBA));
}

ABBA* DCABBA_g_ABBA_ABBA(register ABBA* const this)
{
  DCABBA *tmp;

  print_thunk("DCABBA", "g_ABBA_ABBA", this, -VOFFSET(DCABBA,ABBA), 0);

  tmp = DCABBA_g(DERIVED(this,ABBA,DCABBA));

  print_thunk("DCABBA", "g_ABBA_ABBA", this, VOFFSET(DCABBA,ABBA), 1);

  return BASE(tmp,ABBA);
}

ABBA* DCABBA_h_ABBA_ABBA(register ABBA* const this)
{
  DCABBA *tmp;

  print_thunk("DCABBA", "h_ABBA_ABBA", this, -VOFFSET(DCABBA,ABBA), 0);

  tmp = DCABBA_h(DERIVED(this,ABBA,DCABBA));

  print_thunk("DCABBA", "h_ABBA_ABBA", this, VOFFSET(DCABBA,ABBA), 1);

  return BASE(tmp,ABBA);
}

B* DCABBA_g_BA_B(register BA* const this)
{
  DCABBA *tmp;

  print_thunk("DCABBA", "g_BA_B", this, -VOFFSET(DCABBA,BA), 0);

  tmp = DCABBA_g(DERIVED(this,BA,DCABBA));

  print_thunk("DCABBA", "g_BA_B", this, tmp->vptr[INDEX(DCABBA_B)], 1);

  return VBASE(tmp,DCABBA_B);
}

void DCABBA_f_BA(register BA* const this)
{
  print_thunk("DCABBA", "f_BA", this, -VOFFSET(DCABBA,BA), 0);

  DCABBA_f(DERIVED(this,BA,DCABBA));
}

B* DCABBA_g_B_B(register B* const this)
{
  DCABBA *tmp;

  print_thunk("DCABBA", "g_B_B", this, this->vptr[INDEX(TOP)], 0);

  tmp = DCABBA_g(VTOP(this));

  print_thunk("DCABBA", "g_B_B", this, tmp->vptr[INDEX(DCABBA_B)], 1);

  return VBASE(tmp,DCABBA_B);
}

void DCABBA_f_A(register A* const this)
{
  print_thunk("DCABBA", "f_A", this, this->vptr[INDEX(TOP)], 0);
  
  DCABBA_f(VTOP(this));
}
