#ifndef DCABBA_LAYOUT_H
#define DCABBA_LAYOUT_H

#include "C_layout.h"
#include "D_layout.h"
#include "ABBA_layout.h"

/* -----
 * NOTE: Layout of DCABBA
 * -----
 *
 * struct DCABBA : D, C, ABBA {
 *   DCABBA(long a, long b, long c, long d,
 *          long ab, long ba, long abba, long dcabba_);
 *   ~DCABBA();
 *   void f() const;
 *   DCABBA* g();
 *   DCABBA* h();
 * protected:
 *   long dcabba;
 * };
 *
 * -----
 */

/*
 * Instance layout
 */

typedef struct {
  const VTABLE *vptr;                   /* compiler      */
  long c;                               /* C member      */
  long d;                               /* D member      */
  const VTABLE *ABBA_vptr;              /* compiler      */
  long ab;                              /* AB member     */
  const VTABLE *BA_vptr;                /* compiler      */
  long ba;                              /* BA member     */
  long abba;                            /* ABBA member   */
  long dcabba;                          /* DCABBA member */
  const VTABLE *B_vptr;                 /* compiler      */
  long b;                               /* B member      */
  const VTABLE *A_vptr;                 /* compiler      */
  long a;                               /* A member      */
} DCABBA;

/*
 * Member functions
 */

void    DCABBA_ctor(DCABBA *this,
		    long a, long b, long c, long d,
		    long ab, long ba, long abba, long dcabba);
void    DCABBA_dtor(DCABBA *this);
void    DCABBA_f   (const DCABBA *this);
DCABBA* DCABBA_g   (DCABBA *this);
DCABBA* DCABBA_h   (DCABBA *this);


/* -- -- -- -- -- Compiler extra information -- -- -- -- -- */

/*
 * Virtual bases
 */

enum {
  INDEX(DCABBA_B   ) = INDEX(C_B),
  INDEX(DCABBA_A   ) = AFTER_VBASE(DCABBA_B),
  INDEX(DCABBA_LAST) = INDEX(DCABBA_A)
};

typedef B DCABBA_B;
typedef A DCABBA_A;

/*
 * Virtual member functions
 */

enum {
  INDEX(DCABBA_dtor) = INDEX(C_dtor),
  /* C_g */
  INDEX(DCABBA_f   ) = AFTER_VFUNC(C_g),
  INDEX(DCABBA_g   ) = AFTER_VFUNC(DCABBA_f),
  INDEX(DCABBA_h   ) = AFTER_VFUNC(DCABBA_g)
};

typedef void    (TYPE(DCABBA_dtor))(DCABBA*);
typedef void    (TYPE(DCABBA_f   ))(const DCABBA*);
typedef DCABBA* (TYPE(DCABBA_g   ))(DCABBA*);
typedef DCABBA* (TYPE(DCABBA_h   ))(DCABBA*);


/* -- -- -- -- -- Compiler exported symbols -- -- -- -- -- */

/*
 * Runtime Type Information (RTTI)
 */

extern const type_info DCABBA_info[];

/*
 * Partial constructors and destructors
 */

void DCABBA_ctor_(DCABBA *this, const VTABLE* const vtt[],
		  long a, long b, long c, long d,
		  long ab, long ba, long abba, long dcabba);
void DCABBA_dtor_(DCABBA *this, const VTABLE* const vtt[]);

/*
 * Member functions thunks
 */

void  DCABBA_dtor_top   (DCABBA *this);
B*    DCABBA_g__B       (DCABBA *this);
void  DCABBA_f_ABBA     (ABBA   *this);
ABBA* DCABBA_g_ABBA_ABBA(ABBA   *this);
ABBA* DCABBA_h_ABBA_ABBA(ABBA   *this);
B*    DCABBA_g_BA_B     (BA     *this);
void  DCABBA_f_BA       (BA     *this);
B*    DCABBA_g_B_B      (B      *this);
void  DCABBA_f_A        (A      *this);

#endif  /* DCABBA_LAYOUT_H */
