#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "DCABBA_layout.h"

void dynamic_cast_speed_B_A(B* b)
{ 
  clock_t t0, t1;
  int i;
  A  *a;

  t0 = clock();
  for (i = 0; i < 10000000; i++)
    a = dynamic_cast(b, A);
  t1 = clock();

  printf("dynamic_cast<A*>(b) = %u, ", (size_t)a);
  printf("iter = %ld/sec\n",
	 (long)(10000000/((double)t1-t0)*CLOCKS_PER_SEC));
}

void dynamic_cast_speed_B_C(B* b)
{ 
  clock_t t0, t1;
  int i;
  C  *c;
  
  t0 = clock();
  for (i = 0; i < 10000000; i++)
    c = dynamic_cast(b, C);
  t1 = clock();
  
  printf("dynamic_cast<C*>(c) = %u, ", (size_t)c);
  printf("iter = %ld/sec\n",
	 (long)(10000000/((double)t1-t0)*CLOCKS_PER_SEC));
}

int
main(int argc, char *argv[])
{
  int do_flags = argc > 1 ? strtol(argv[1],0,0) : 0;
  int do_layout = do_flags & 1;
  int do_thunk = do_flags & 2;

  A      a     [1];
  B      b     [1];
  C      c     [1];
  AB     ab    [1];
  BA     ba    [1];
  ABBA   abba  [1];
  D      d     [1];
  DCABBA dcabba[1];

  print_thunk_flag (do_thunk);
  print_layout_flag(do_layout);

  do_layout && printf("*** A Layout\n");
  A_ctor(a,1);    

  do_layout && printf("*** B Layout\n");
  B_ctor(b,2);    

  do_layout && printf("*** C Layout\n");
  C_ctor(c,2,3);  

  do_layout && printf("*** D Layout\n");
  D_ctor(d,4);  

  do_layout && printf("*** AB Layout\n");
  AB_ctor(ab,1,2,5);

  do_layout && printf("*** BA Layout\n");
  BA_ctor(ba,1,2,6);

  do_layout && printf("*** ABBA Layout\n");
  ABBA_ctor(abba,1,2,5,6,7);

  do_layout && printf("*** DCABBA Layout\n");
  DCABBA_ctor(dcabba,1,2,3,4,5,6,7,8);

  printf("sizeof a      = %2d @ %u\n", sizeof a     , (size_t)a     );
  printf("sizeof b      = %2d @ %u\n", sizeof b     , (size_t)b     );
  printf("sizeof c      = %2d @ %u\n", sizeof c     , (size_t)c     );
  printf("sizeof d      = %2d @ %u\n", sizeof d     , (size_t)d     );
  printf("sizeof ab     = %2d @ %u\n", sizeof ab    , (size_t)ab    );
  printf("sizeof ba     = %2d @ %u\n", sizeof ba    , (size_t)ba    );
  printf("sizeof abba   = %2d @ %u\n", sizeof abba  , (size_t)abba  );
  printf("sizeof dcabba = %2d @ %u\n", sizeof dcabba, (size_t)dcabba);

  VCALL0(a, A_f);                       /* a.f(); */
  VCALL0(b, B_g);                       /* b.g(); */

  VCALL0(c, C_g);                       /* c.g(); */
  VCALL0(BASE(c,B), B_g);               /* static_cast<B*>(&c)->g(); */

  VCALL0(ab, AB_f);                     /* ab.f(); */
  VCALL0(ab, AB_g);                     /* ab.g(); */
  VCALL0(ab, AB_h);                     /* ab.h(); */
  VCALL0(BASE(ab,A), A_f);              /* static_cast<A*>(&ba)->f(); */
  VCALL0(BASE(ab,B), B_g);              /* static_cast<B*>(&ba)->g(); */
  
  VCALL0(ba, BA_f);                     /* ba.f(); */
  VCALL0(ba, BA_g);                     /* ba.g(); */
  VCALL0(BASE(ba,A), A_f);              /* static_cast<A*>(&ba)->f(); */
  VCALL0(BASE(ba,B), B_g);              /* static_cast<B*>(&ba)->g(); */

  VCALL0(abba, ABBA_f);                 /* abba.f(); */
  VCALL0(abba, ABBA_g);                 /* abba.g(); */
  VCALL0(abba, ABBA_h);                 /* abba.h(); */
  VCALL0((AB*)abba, AB_f);              /* static_cast<AB*>(&abba)->f(); */
  VCALL0((AB*)abba, AB_g);              /* static_cast<AB*>(&abba)->g(); */
  VCALL0((AB*)abba, AB_h);              /* static_cast<AB*>(&abba)->g(); */
  VCALL0(BASE(abba,BA), BA_f);          /* static_cast<BA*>(&abba)->f(); */
  VCALL0(BASE(abba,BA), BA_g);          /* static_cast<BA*>(&abba)->g(); */
  VCALL0(BASE(abba,A), A_f);            /* static_cast<A*>(&abba)->f();  */
  VCALL0(BASE(abba,B), B_g);            /* static_cast<B*>(&abba)->g();  */

  VCALL0(dcabba, DCABBA_f);             /* dcabba.f(); */
  VCALL0(dcabba, DCABBA_g);             /* dcabba.g(); */
  VCALL0(dcabba, DCABBA_h);             /* dcabba.h(); */
  VCALL0((C*)dcabba, C_g);              /* static_cast<C*>(&dcabba)->g(); */
  VCALL0(BASE(dcabba,ABBA), ABBA_f);    /* static_cast<ABBA*>(&dcabba)->f(); */
  VCALL0(BASE(dcabba,ABBA), ABBA_g);    /* static_cast<ABBA*>(&dcabba)->g(); */
  VCALL0(BASE(dcabba,ABBA), ABBA_h);    /* static_cast<ABBA*>(&dcabba)->h(); */
  VCALL0((AB*)BASE(dcabba,ABBA), AB_f); /* ((AB*)&dcabba)->f(); */
  VCALL0((AB*)BASE(dcabba,ABBA), AB_g); /* ((AB*)&dcabba)->g(); */
  VCALL0((AB*)BASE(dcabba,ABBA), AB_h); /* ((AB*)&dcabba)->h(); */
  VCALL0(BASE(dcabba,BA), BA_f);        /* static_cast<BA*>(&dcabba)->f();*/
  VCALL0(BASE(dcabba,BA), BA_g);        /* static_cast<BA*>(&dcabba)->g(); */
  VCALL0(BASE(dcabba,A), A_f);          /* static_cast<A*>(&dcabba)->f();  */
  VCALL0(BASE(dcabba,B), B_g);          /* static_cast<B*>(&dcabba)->g();  */

  printf("dynamic_cast<AB*>((B*)dcabba) = %u\n",
	 (size_t)dynamic_cast(BASE(dcabba,B), AB));

  printf("dynamic_cast<BA*>((B*)dcabba) = %u\n",
	 (size_t)dynamic_cast(BASE(dcabba,B), BA));

  printf("dynamic_cast<C*>((B*)dcabba) = %u\n",
	 (size_t)dynamic_cast(BASE(dcabba,B), C));

  printf("AB    : ");
  dynamic_cast_speed_B_A(BASE(ab,B));
  printf("DCABBA: ");
  dynamic_cast_speed_B_A(BASE(dcabba,B));
  printf("DCABBA: ");
  dynamic_cast_speed_B_C(BASE(dcabba,B));

  do_layout && printf("*** A Layout\n");
  DCABBA_dtor(dcabba);

  do_layout && printf("*** B Layout\n");
  ABBA_dtor(abba);

  do_layout && printf("*** C Layout\n");
  BA_dtor(ba);

  do_layout && printf("*** D Layout\n");
  AB_dtor(ab);
  
  do_layout && printf("*** AB Layout\n");
  D_dtor(d);

  do_layout && printf("*** BA Layout\n");
  C_dtor(c);

  do_layout && printf("*** ABBA Layout\n");
  B_dtor(b);

  do_layout && printf("*** DCABBA Layout\n");
  A_dtor(a);

  return 0;
}
