#include "object_model.h"

typedef struct {
  const VTABLE *vptr;
} OBJECT;

/*
 * Private functions
 */

static int
base_cast(OBJECT *this, const type_info *target_type,
          const type_info *type, void **found)
{
  const struct base_info *base;
  unsigned i, cnt;

  base = type->base;
  cnt  = type->base_cnt;

  for (i = 0; i < cnt; i++) {

    if (base[i].base_flag & virtual_base)
      break;

    if (base[i].base_flag & public_base) {
      void *sub = (char*)this + base[i].base_offset;

      if (base[i].base_type == target_type) {
        if(*found)
          return *found = 0, 1;
        found = sub;
      }

      if (base[i].base_type->base_cnt)
        if (base_cast(sub, target_type, base[i].base_type, found))
          return 1;
    }
  }

  return 0;
}

/*
 * Public functions
 */

const type_info* (typeid)(void *this_)
{
  OBJECT *this = this_;

  if (this)
    return (void*)this->vptr[INDEX(TINFO)];

  return 0; /* throw bad_typeid; */
}

void* (dynamic_cast)(void *this_, const type_info *target_type)
{
  const struct base_info *base;
  const type_info *type;
  OBJECT *this = this_;
  unsigned i, cnt;
  void *found;

  if (!this) return 0;

  type = typeid(this);

  if (type == target_type)
    return this;

  this  = VTOP(this);
  base  = type->base;
  cnt   = type->base_cnt;
  found = 0;

  for (i = 0; i < cnt; i++) {

    if (base[i].base_flag & public_base) {
      void *sub = (char*)this + base[i].base_offset;

      if (base[i].base_type == target_type) {
        if(found) return 0;
        found = sub;
      }

      if (base[i].base_type->base_cnt)
        if (base_cast(sub, target_type, base[i].base_type, &found))
          return found;
    }
  }

  return found;
}
