#include <stdio.h>
#include <stdarg.h>
#include "object_model.h"

static int do_print_layout = 0;
static int do_print_thunk  = 0;

void
print_layout_flag(int do_it)
{
  do_print_layout = do_it;
}

void
print_thunk_flag(int do_it)
{
  do_print_thunk = do_it;
}

void
print_layout(const char *cls, const char *func,
	     void *this, size_t size)
{
  int i, cnt = size/sizeof(long);
  long *p = this;

  if (!do_print_layout)
    return;

  if (size % sizeof(long) > 0) {
    fprintf(stderr, "\tinvalid integral type 'long' for layout\n");
    return;
  }

  printf("%s::%s layout @ %u : %d bytes\n", cls, func, (size_t)p, size);

  for (i = 0; i < cnt; i++)
    printf("\t[%d]=%ld\n", i, p[i]);
}

void
print_thunk(const char *cls, const char *func,
	    void *this, int offset, int is_ret)
{
  if (do_print_thunk)
    printf("thunk %6s::%8s(%10u)\t[%s %+d]\n", cls, func,
	   (size_t)this, is_ret ? "<-" : "->", offset);
}

void
print_vfunc(const char *cls, const char *func, const void *this,
	    const char *fmt, ...)
{
  va_list va;

  if (do_print_thunk)
    printf("vcall %6s::%8s(%10u)\t", cls, func, (size_t)this);
  else
    printf("%6s::%s(%10u)\t", cls, func, (size_t)this);
    
  va_start(va, fmt);
  vprintf(fmt, va);
  va_end(va);
  puts("");
}

void
print_mfunc(const char *cls, const char *func, const void *this,
	    const char *fmt, ...)
{
  va_list va;

  if (do_print_thunk)
    printf("mcall %6s::%8s(%10u)\t", cls, func, (size_t)this);
  else
    printf("%6s::%s(%10u)\t", cls, func, (size_t)this);
    
  va_start(va, fmt);
  vprintf(fmt, va);
  va_end(va);
  puts("");
}
